/*-------------------------------------------------------------
 *  Copyright (c) 2011-2014, MSS ComBlock. All rights reserved.
 *
 * 	Filename: RF.c
 *  Last modified: 8/9/14
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 5
 *  Inheritance: N/A
 *  Description: RF components driver
 *
 *  Rev 1.1 6/1/11 AZ
 *  Moved RF gain control 2 from PWM 1.2 to PWM 1.4 (reason: hardware version change from prototype to production)
 *
 *  Rev 2 4/25/12 AZ
 *  Corrected bug in program_tuner_pll
 *
 *  Rev 3 3/9/13 AZ
 *  Read mixer power detector, even when AGC loop is open
 *  reduced AGC threshold to prevent saturation (wideband noise + signal)
 *  Note: make sure shield is soldered on all 4 edges to prevent oscillations.
 *
 *  Rev 4 4/15/14 AZ
 *  Rewrote AD7298_rw to shave a few us (3 us).
 *  Added ISRs triggered by falling and rising edges of RX_TXN to speed up the switching
 *
 *  Rev 5 8/9/14 AZ
 *  Reduced the rx RF synthesizer level from -1 to -4 dBm to minimize leakage (will be fixed in
 *  next hardware rev)
 *  Changed the RF AGC 'too hot' condition: RF gain is reduced when power detected
 *  at the mixer exceeds the threshold (normal) AND IF power detected too small.
 *  This is to prevent AGC deadlock when LO leakage is in-band and exceeds rf signal level.
 -------------------------------------------------------------*/
#include "main.h"
#include "rf.h"
#include "utils.h"

/*-------- change this section according to the schematics ----------------------------------------*/
inline long synth_check_lock(int synth_no) {
	if(synth_no == 1){
		return FIO1PIN & 0x00000100; // pin P1[8] ADF4351 synthesizer lock status
	} else if (synth_no == 2){
		return FIO1PIN & 0x00040000; // pin P1[18] ADF4351 synthesizer lock status
	}
}


/*-------- end of schematics-dependent code ----------------------------------------*/



unsigned long get_freq(long index) {
	// reconstruct the RF frequency synthesizer frequency from the control registers.
	// index range 0 - 7
	// Returns the frequency expressed in Hz.
	long reg_no;
	unsigned long freq;
	long i;

	// points to control register with MSB
	if (index == 0)
		reg_no = 3; // frequency in REG3-0
	else if (index < 8)
		reg_no = 6 + 4 * index;
	else
		return 0; // invalid index

	for (i = 0; i < 4; i++) {
		freq <<= 8;
		freq &= 0xFFFFFF00;
		freq |= ((unsigned long) regv[reg_no--]) & 0x000000FF;
	}
	return freq;
}

void adf4351_reg(long regval, int synth_no) {
	// write 32 serial bits to the ADF4351 Integrated synthesizer and VCO

	int i;
	if(synth_no == 1){
		PIN_SYNTH1_CLK_L(); // always start transaction with clock low
		PIN_SYNTH1_LE_L(); // Latch enable - low  (initialized high at power up)

		for (i = 0; i < 32; i++) {
			if (regval & 0x80000000) {
				PIN_SYNTH1_DATA_H(); //data high
				PIN_SYNTH1_CLK_H(); //clock high
			} else {
				PIN_SYNTH1_DATA_L(); //data low
				PIN_SYNTH1_CLK_H(); //clock high
			}

			regval <<= 1;
			PIN_SYNTH1_CLK_L(); //clock low
		}

		PIN_SYNTH1_LE_H(); // Latch disable - high
	} else if(synth_no == 2){
		PIN_SYNTH2_CLK_L(); // always start transaction with clock low
		PIN_SYNTH2_LE_L(); // Latch enable - low  (initialized high at power up)

		for (i = 0; i < 32; i++) {
			if (regval & 0x80000000) {
				PIN_SYNTH2_DATA_H(); //data high
				PIN_SYNTH2_CLK_H(); //clock high
			} else {
				PIN_SYNTH2_DATA_L(); //data low
				PIN_SYNTH2_CLK_H(); //clock high
			}

			regval <<= 1;
			PIN_SYNTH2_CLK_L(); //clock low
		}

		PIN_SYNTH2_LE_H(); // Latch disable - high

	}
}

//-----------------------------------------------------------------------
//-- RF synthesizer profiles
//
// init_tuner_pll(,1);  // 26 MHz frequency reference
// init_tuner_pll(,2);  // 19.2 MHz frequency reference, integer-N
// init_tuner_pll(,3);  // 19.2 MHz frequency reference, fractional-N (modified from COM-3011 with 20 MHz frequency reference)
// init_tuner_pll(,4);  // 10 MHz frequency reference, fractional-N


void init_tuner_pll(int synth_no, char profile_no) {
	// one-time ADF4351 initialization.
	// program registers in sequence 5,4,3,2,1,0
	unsigned long regval;

	// register R5
	regval = 0x00580005;
	// LD pin = digital lock detect
	adf4351_reg(regval, synth_no);

	if(profile_no == 1){
		// 26 MHz frequency reference. Pfd = 13 MHz. Integer-N

		// register R3
		regval = 0x00E48283;
		// clock divider value = 50
		// fast lock enable
		// cycle slip reduction enable
		// charge cancellation
		// band select clock mode high
		adf4351_reg(regval, synth_no);

		// register R2
	//	regval = 0x1806AFC2;   // R counter = 26, reference doubler disabled, reference divide by 2 disabled
		regval = 0x19036FC2;   // R counter = 13, reference doubler disabled, reference divide by 2 enabled


		// counter reset disabled
		// cp three-state disabled
		// not powered down
		// pd positive polarity, 10ns LDP
		// int-N, icp = 2.5mA
		// double buffer enabled
		// muxout = digital lock detect
		// low phase noise mode
		adf4351_reg(regval, synth_no);

		// register R1
		regval = 0x00008011; // 4/5 prescaler,
		adf4351_reg(regval, synth_no);
	} else if (profile_no == 2){
		// 19.2 MHz frequency reference. Pfd = 400 kHz. Integer-N

		// register R3
		regval = 0x00E08023;
		// clock divider value = 4
		// fast lock enable
		// cycle slip reduction disable
		// charge cancellation
		// band select clock mode high
		adf4351_reg(regval, synth_no);

		// register R2
		regval = 0x1B0C21C2;   // R counter = 48, reference doubler enabled, reference divide by 2 enabled
		// counter reset disabled
		// cp three-state disabled
		// not powered down
		// pd positive polarity, 6ns LDP
		// int-N, icp = 0.31mA
		// double buffer enabled
		// muxout = digital lock detect
		// low phase noise mode
		adf4351_reg(regval, synth_no);

		// register R1
		regval = 0x08008011; // 8/9 prescaler,
		adf4351_reg(regval, synth_no);

	} else if (profile_no == 3){
		// 19.2 MHz frequency reference. Pfd = 5 kHz. Fractional-N

		// register R3
		regval = 0x00008013;
		// clock divider value = 400uS * 5 KHz = 2
		// fast clock enable
		// cycle slip reduction disabled (seems to increase phase noise a lot! when enabled)
		adf4351_reg(regval, synth_no);

		// register R2
		// double 19.2 MHz ref clock for better phase noise)
		regval = 0x18006E42; //  preference: low noise mode
		// counter reset disabled
		// cp three-state disabled
		// not powered down
		// pd positive polarity, 10ns LDP
		// frac-N, icp = 2.5ma
		// double buffer enabled
		// 10-bit R counter = 1, NO ref doubling NO 1:2
		// muxout = digital lock detect
		// low phase noise mode
		adf4351_reg(regval, synth_no);

		// register R1
		regval = 0x0800F801; // 8/9 prescaler, MOD = 3840
		adf4351_reg(regval, synth_no);
	} else if (profile_no == 4){
		// 10 MHz frequency reference. Pfd = 5 kHz. Fractional-N

		// register R3
		regval = 0x00008013;
		// clock divider value = 400uS * 5 KHz = 2
		// fast clock enable
		// cycle slip reduction disabled (seems to increase phase noise a lot! when enabled)
		adf4351_reg(regval, synth_no);

		// register R2
		// double 10 MHz ref clock for better phase noise)
		regval = 0x18006E42; //  preference: low noise mode
		// counter reset disabled
		// cp three-state disabled
		// not powered down
		// pd positive polarity, 10ns LDP
		// frac-N, icp = 2.5ma
		// double buffer enabled
		// 10-bit R counter = 1, NO ref doubling NO 1:2
		// muxout = digital lock detect
		// low phase noise mode
		adf4351_reg(regval, synth_no);

		// register R1
		regval = 0x0800BE81; // 8/9 prescaler, MOD = 2000
		adf4351_reg(regval, synth_no);
	}
	return;
}

void program_tuner_pll(int synth_no, char profile_no) {
	// program RF synthesizer with default frequency
	// One must call init_tuner_pll() once prior to using this function (one-time).

	unsigned long regval, f, fvcodiv2, fi;
	unsigned short nint, nfrac;

	// construct frequency word (in Hz) from control registers
	if(synth_no == 1) {
		// Rx frequency index is in REG6[3:0]
		fi = (regv[6] & 0x0000000F);
		// receive frequency to IF_CENTER_FREQ IF
		f = get_freq(fi); // frequency in Hz
		f -= IF_CENTER_FREQ;	// IF center frequency
	} else if(synth_no == 2){
		// baseband to transmit frequency
		// Tx frequency index is in REG6[7:4]
		fi = (regv[6] & 0x000000F0) >> 4;
		f = get_freq(fi); // frequency in Hz
	}



	if(profile_no == 1){
		// 26 MHz frequency reference. Pfd = 13 MHz. Integer-N

		// program registers in sequence 5,4,3,2,1,0
		// register R4
		// convert tuner_freq into parameters understandable by the ADF4351
		// multiply turn tuner_freq to get a VCO frequency before the final output divider
		// Disable auxiliary output (RF_OUTB unused)
		// Set output LO power to -1dBm (trying to minimize LO leakage to the RF input).
		if (f >= 2200000000) {
			regval = 0x0080842C; // fout >= 2.2 GHz
			fvcodiv2 = f >> 1; // VCO frequency/2 (to prevent overflow later in the computation)
		} else if (f >= 1100000000) {
			regval = 0x0090842C; // fout >= 1.1 GHz
			fvcodiv2 = f;
		} else if (f >= 550000000) {
			regval = 0x00A0842C; // fout >= 550 MHz
			fvcodiv2 = f << 1;
		} else if (f >= 275000000) {
			regval = 0x00B0842C; // fout >= 275 MHz
			fvcodiv2 = f << 2;
		} else if (f >= 137500000) {
			regval = 0x00C0842C; // fout >= 137.5 MHz
			fvcodiv2 = f << 3;
		} else if (f >= 68750000) {
			regval = 0x00D0842C; // fout >= 68.75 MHz
			fvcodiv2 = f << 4;
		} else {
			regval = 0x00E0842C; // fout >= 34.375 MHz
			fvcodiv2 = f << 5;
		}
		// feedback select = fundamental
		// band select clock divider = 26MHz/125KHz max = 208. use 220 (xDC).
		// vco powered up, mute till lock detect,
		// aux output disabled, divided outputs
		adf4351_reg(regval, synth_no);

		// register R0
		// reference frequency is 1 MHz
		nint = (unsigned short) (fvcodiv2 / (1000000/2));
		// integer-N PLL
		regval = (unsigned long) (nint);
		regval <<= 15;
		regval = regval & 0x7FFF8000;
		adf4351_reg(regval, synth_no);
	} else if (profile_no == 2){
		// 19.2 MHz frequency reference. Pfd = 200 kHz. Integer-N

		// program registers in sequence 5,4,3,2,1,0
		// register R4
		// convert tuner_freq into parameters understandable by the ADF4351
		// multiply turn tuner_freq to get a VCO frequency before the final output divider
		// Disable auxiliary output (RF_OUTB unused)
		// Set output LO power to -1dBm (trying to minimize LO leakage to the RF input).
		if (f >= 2200000000) {
			regval = 0x0080842C; // fout >= 2.2 GHz
			fvcodiv2 = f >> 1; // VCO frequency/2 (to prevent overflow later in the computation)
		} else if (f >= 1100000000) {
			regval = 0x0090842C; // fout >= 1.1 GHz
			fvcodiv2 = f;
		} else if (f >= 550000000) {
			regval = 0x00A0842C; // fout >= 550 MHz
			fvcodiv2 = f << 1;
		} else if (f >= 275000000) {
			regval = 0x00B0842C; // fout >= 275 MHz
			fvcodiv2 = f << 2;
		} else if (f >= 137500000) {
			regval = 0x00C0842C; // fout >= 137.5 MHz
			fvcodiv2 = f << 3;
		} else if (f >= 68750000) {
			regval = 0x00D0842C; // fout >= 68.75 MHz
			fvcodiv2 = f << 4;
		} else {
			regval = 0x00E0842C; // fout >= 34.375 MHz
			fvcodiv2 = f << 5;
		}
		// feedback select = fundamental
		// band select clock divider = 26MHz/125KHz max = 208. use 220 (xDC).
		// vco powered up, mute till lock detect,
		// aux output disabled, divided outputs
		adf4351_reg(regval, synth_no);

		// register R0
		// reference frequency is 400 kHz
		nint = (unsigned short) (fvcodiv2 / (400000/2));
		// integer-N PLL
		regval = (unsigned long) (nint);
		regval <<= 15;
		regval = regval & 0x7FFF8000;
		adf4351_reg(regval, synth_no);

	} else if (profile_no == 3){
		// 19.2 MHz frequency reference. Pfd = 5 kHz. Fractional-N

		// register R4
		// convert tuner_freq into parameters understandable by the ADF4351
		// multiply turn tuner_freq to get a VCO frequency before the final output divider
		// Disable auxiliary output (RF_OUTB unused)
		// Set output LO power to -1dBm (trying to minimize LO leakage to the RF input).
		if (f >= 2200000000) {
			if(synth_no == 1) {
				// Rx. Reduce LO level to -4 dBm(see DR217)
				regval = 0x008C8424; // fout >= 2.2 GHz
			} else {
				regval = 0x008C842C; // fout >= 2.2 GHz
			}
			fvcodiv2 = f >> 1; // VCO frequency/2 (to prevent overflow later in the computation)
		} else if (f >= 1100000000) {
			if(synth_no == 1) {
				// Rx. Reduce LO level to -4 dBm(see DR217)
				regval = 0x009C8424; // fout >= 2.2 GHz
			} else {
				regval = 0x009C842C; // fout >= 1.1 GHz
			}
			fvcodiv2 = f;
		} else if (f >= 550000000) {
			regval = 0x00AC842C; // fout >= 550 MHz
			fvcodiv2 = f << 1;
		} else if (f >= 275000000) {
			regval = 0x00BC842C; // fout >= 275 MHz
			fvcodiv2 = f << 2;
		} else if (f >= 137500000) {
			regval = 0x00CC842C; // fout >= 137.5 MHz
			fvcodiv2 = f << 3;
		} else if (f >= 68750000) {
			regval = 0x00DC842C; // fout >= 68.75 MHz
			fvcodiv2 = f << 4;
		} else {
			regval = 0x00EC842C; // fout >= 34.375 MHz
			fvcodiv2 = f << 5;
		}
		// feedback select = fundamental
		// band select clock divider = 19.2MHz/125KHz max = 154. use 200 (xC8).
		// vco powered up, mute till lock detect,
		// aux output disabled, divided outputs
		adf4351_reg(regval, synth_no);

		// register R0
		// reference frequency is 19.2 MHz
		nint = (unsigned short) (fvcodiv2 / 9600000);
		fvcodiv2 -= nint * 9600000;
		// frac step size is 5KHz
		nfrac = (unsigned short) (fvcodiv2 * 2 / 5000);
		regval = (unsigned long) (nint);
		regval <<= 12;
		regval = regval & 0x0FFFF000;
		regval |= ((unsigned long) (nfrac)) & 0x000FFF;
		regval <<= 3;
		regval &= 0x7FFFFFF8;
		adf4351_reg(regval, synth_no);

	} else if (profile_no == 4){
		// 10 MHz frequency reference. Pfd = 5 kHz. Fractional-N

		// register R4
		// convert tuner_freq into parameters understandable by the ADF4351
		// multiply turn tuner_freq to get a VCO frequency before the final output divider
		// Disable auxiliary output (RF_OUTB unused)
		// Set output LO power to -1dBm (trying to minimize LO leakage to the RF input).
		if (f >= 2200000000) {
			regval = 0x008C842C; // fout >= 2.2 GHz
			fvcodiv2 = f >> 1; // VCO frequency/2 (to prevent overflow later in the computation)
		} else if (f >= 1100000000) {
			regval = 0x009C842C; // fout >= 1.1 GHz
			fvcodiv2 = f;
		} else if (f >= 550000000) {
			regval = 0x00AC842C; // fout >= 550 MHz
			fvcodiv2 = f << 1;
		} else if (f >= 275000000) {
			regval = 0x00BC842C; // fout >= 275 MHz
			fvcodiv2 = f << 2;
		} else if (f >= 137500000) {
			regval = 0x00CC842C; // fout >= 137.5 MHz
			fvcodiv2 = f << 3;
		} else if (f >= 68750000) {
			regval = 0x00DC842C; // fout >= 68.75 MHz
			fvcodiv2 = f << 4;
		} else {
			regval = 0x00EC842C; // fout >= 34.375 MHz
			fvcodiv2 = f << 5;
		}
		// feedback select = fundamental
		// band select clock divider = 10MHz/125KHz max = 80. use 100.
		// vco powered up, mute till lock detect,
		// aux output disabled, divided outputs
		adf4351_reg(regval, synth_no);

		// register R0
		// reference frequency is 10 MHz
		nint = (unsigned short) (fvcodiv2 / 5000000);
		fvcodiv2 -= nint * 5000000;
		// frac step size is 5KHz
		nfrac = (unsigned short) (fvcodiv2 * 2 / 5000);
		regval = (unsigned long) (nint);
		regval <<= 12;
		regval = regval & 0x0FFFF000;
		regval |= ((unsigned long) (nfrac)) & 0x000FFF;
		regval <<= 3;
		regval &= 0x7FFFFFF8;
		adf4351_reg(regval, synth_no);
	}
}

inline void RX_TXN_R_ISR(){
	// Interrupt service routine when the RX_TX# signal switches to 1
	// called at rising edge.
	// half-duplex operation, receive enabled, tx disabled
	DAC(0);	// minimize TX_GAIN_CTRL2. Do this first (slow)
	// enable/disable RF synthesizer output depending on the RX_TXN signal
	PIN_SYNTH1_ENABLE_H();
	PIN_SYNTH2_ENABLE_L();
	PIN_RFOUT_ENB_H();	// disable the RF modulator output (active low)
}
inline void RX_TXN_F_ISR(){
	// Interrupt service routine when the RX_TX# signal switches to 0
	// called at falling edge.
	// half-duplex operation, tx enabled, rx disabled
	setTxGain();	// do this first (slow)
	// enable Tx RF synthesizer
	PIN_SYNTH2_ENABLE_H();
	// tx RF gain control
	PIN_RFOUT_ENB_L();	// enable the RF modulator output (active low)
	// disable Rx RF synthesizer to minimize RFI
	PIN_SYNTH1_ENABLE_L();
}

// call once at power up or after change in configuration
void RX_TXN_poll() {
	// only call while in half-duplex mode.
	if(FIO0PIN & 0x02000000){
		// receive enabled, tx disabled
		// enable/disable RF synthesizer output depending on the RX_TXN signal
		PIN_SYNTH1_ENABLE_H();
		PIN_SYNTH2_ENABLE_L();
		PIN_RFOUT_ENB_H();	// disable the RF modulator output (active low)
		DAC(0);	// minimize TX_GAIN_CTRL2
	} else {
		// tx enabled, rx disabled
		// enable/disable RF synthesizer output depending on the RX_TXN signal
		PIN_SYNTH1_ENABLE_L();
		PIN_SYNTH2_ENABLE_H();
		PIN_RFOUT_ENB_L();	// enable the RF modulator output (active low)
		// tx RF gain control
		setTxGain();
	}
}

// check TX_EN input on pin A17 (P1[28]). Control the RF_OUT_EN at the quadrature modulator.
inline void TX_EN_poll() {
	// LATENCY: 0 - 9us
	if(FIO1PIN & 0x10000000){
		PIN_RFOUT_ENB_L();	// enable the RF modulator output (active low)
		PIN_SYNTH2_ENABLE_H();	// enable the RF synthesizer output
	} else {
		PIN_RFOUT_ENB_H();	// disable the RF modulator output
		PIN_SYNTH2_ENABLE_L();	// disable the RF synthesizer output
	}
}

void DAC_init(void) {
	// initialize internal 10-bit DAC

	// 1. to enable the DAC, output must be selected to appear on the related pin p0.26
	PINSEL1 = (PINSEL1 & 0xFFCFFFFF) | 0x00200000;

	// 2. peripheral clock selection for DAC.
	// select the processor clock CCLK/4 (default)
	PCLKSEL1 = PCLKSEL1 & 0xFF3FFFFF;

	// 3. enable double buffering, no interrupts, 1us refresh
	DACCTRL = 0x00000006;
	DACCNTVAL = 30; // 1us @120MHz/4
}

void DAC(long value) {
	// output is value * Vrefp / 1024
	// 1us settling time -> 1 MHz maximum update rate
	DACR = (value << 6) & 0x0000FFC0;
}

void PWM_init(void) {
	// initialized P1[26] = PWM 1.6 as a fake 10-bit DAC (externally filtered)
	// PWM1IR interrupt register ignored (no interrupts)
	PCONP |= (1 << 6); // PWM1 power ON
	// PCLKSEL0: by default peripheral clock is CCLK/4 = 30 MHz
	PWM1TCR = 0x00000002; // Reset counter
	PWM1PR = 0;
	PWM1PC = 0;
	PINSEL3 = (PINSEL3 & 0xFFCFFFFF) | 0x00200000;
	// p1[26] = PWM1.6
	PINMODE3 = (PINMODE3 & 0xFFCFFFFF) | 0x00200000;
	// P1[26] PWM has no PU/PD
	PWM1MR0 = 1024; // controls the PWM cycle rate. Here 1024 peripheral clocks
	PWM1MR6 = 0; // initial PWM value
	// PWM Match Control Register
	// Match 0 is configured to reset the timer/counter when a match event occurs.
	PWM1MCR = 0x00000002; // reset counter when match MR0
	// PWM Control Register
	PWM1PCR = 0x00004000; // The PWM6 output enabled
	// PWM Timer Control Register
	PWM1TCR = 0x00000009; // clear counter,  enable PWM mode, enable counting
}

// control PWM1.6
void PWM(long value) {
	// output is value * 3.3V / 1024 after filtering
	PWM1MR0 = 1024; // controls the PWM cycle rate. Here 1024 peripheral clocks
	PWM1MR6 = value & 0x000003FF; // controls the PWM edge position

	// PWM Latch Enable Register
	PWM1LER = 0x00000041; // Enable PWM Match 0 and 6 Latch
}

void setLNAgain(void){
	PWM(LNA_gain);
}

void setRFGain(void) {
	AD5621_1(RF_gain);
}

void setIFGain(void) {
	AD5621_2(IF_gain);

}

// call once after tx is enabled
inline void setTxGain(void){
	int c;
	if((regv[39] & 0x30) == 0){
		// tx gain set by control register. fixed
		// transmitter gain (10-bit)
		c = (int)(regv[38]) & 0x00000003;
		c <<= 8;
		c |= (int)(regv[37]) & 0x000000FF;
		DAC(c);
	} else if ((regv[39] & 0x30) == 0x10){
		// tx gain set by external analog signal TX_GAIN_CTRL1
		c = aux_adc[5];
//		c = (c >> 2) & 0x03FF;	// reduce to 10-bit
		c >>= 2;	// reduce to 10-bit
		DAC(c);
	}
}

// call periodically to track TX_GAIN_CTRL1
void tx_gain_poll(void){
	int c;
	if ((regv[39] & 0x30) == 0x10){
		// tx gain set by external analog signal TX_GAIN_CTRL1
		c = aux_adc[5];
//		c = (c >> 2) & 0x03FF;	// reduce to 10-bit
		c >>= 2;	// reduce to 10-bit
		DAC(c);
	}
}

void LNA_AGC_poll(void){
	// adjust the 1st attenuator after the LNA to prevent saturation,
	// based on the RSSI measurement

	// 2 AGC modes as set in REG39(0)
	unsigned char agc_loop;
	agc_loop = regv[39] & 0x01;


	if (agc_loop == 0) {
		// open loop. Gain is fixed.
		return;
	} else if (agc_loop == 1){
		// LNA AGC closed loop algorithm relying on RSSI (prevent saturation at
		// amplifiers), RSSI (RF_POWER_DET1) -> aux_adc[0]

		// response: flat at 1023 from 0 to 0x800, then linear down to 0 at 0xFFF
		// based on COM-0013 board measurements.
		if(aux_adc[0] > 0x0800){
			LNA_gain = 1023 - (((aux_adc[0] - 0x800) >> 1)& 0x3FF);
		} else {
			LNA_gain = 1023;
		}
		setLNAgain();	// set LNA gain
	}
}

void RF_AGC_poll(void) {
	// 2 AGC modes as set in REG39(1:0)
	unsigned char agc_loop;
	agc_loop = (regv[39] >> 1) & 0x01;

	// check level/power sensors for reporting status (even when AGC loop is open) 3/9/13
	if (agc_loop == 0) {
		// open loop. Gain is fixed.
		return;
	} else if(((regv[39] & 0x0040) == 0) && ((FIO1PIN & 0x20000000) == 0)){
		// freeze RF AGC when in half-duplex mode and RX_TXN = '0'
		return;
	} else if (agc_loop == 1){
		// RF AGC closed loop algorithm relying on RF power detection (prevent saturation at
		// RF mixer), RF_POWER_DET2 -> aux_adc[1]

		// read RF power detect rf_power_det (done periodically in main loop)
		// new: threshold reduced to prevent saturation (wideband noise + signal)
//		if ((aux_adc[1] > 0x0FA8) && (aux_adc[2] > 0x0E80)) {
//		if (aux_adc[1] > 0x0E74) {
//  		simpler but may deadlock if out-of-band signal or LO leakage is stronger than rx signal

		if ((aux_adc[1] > 0x0E74) && (aux_adc[2] < 0x0010)) {
			// special case: in-band LO leakage is too strong (will be corrected in next
			// hardware rev). Do not decrease RF gain. On the contrary, increase it so as
			// not to supress the rx signal
			// slow pressure to increase gain.
			RF_gain_too_small_duration++; // increment the time elapsed since last saturation detect
			if (RF_gain_too_small_duration > 512) { // agc reaction time implied here
				// gain is consistently too small.
				RF_gain_increment();
				RF_gain_too_large_duration = 0; // reset counter
				RF_gain_too_small_duration = 0; // reset the time elapsed since last saturation detect
			}

		} else if (aux_adc[1] > 0x0E74) {
			// getting close to saturation at the RF mixer. Decrease gain
			RF_gain_too_large_duration++;
			RF_gain_too_small_duration = 0; // reset the time elapsed since last peak detect > 0x80
			if (RF_gain_too_large_duration > 512) { // agc reaction time implied here
				RF_gain_too_large_duration = 0; // reset counter
				RF_gain_decrement();
			}
		} else {
			// slow pressure to increase gain.
			RF_gain_too_small_duration++; // increment the time elapsed since last saturation detect
			if (RF_gain_too_small_duration > 512) { // agc reaction time implied here
				// gain is consistently too small.
				RF_gain_increment();
				RF_gain_too_large_duration = 0; // reset counter
				RF_gain_too_small_duration = 0; // reset the time elapsed since last saturation detect
			}
		}
	}
}

void IF_AGC_poll(void) {
	// 3 AGC modes as set in REG39(3:2)
	unsigned char agc_loop;
	agc_loop = (regv[39] & 0x0C) >> 2;

	if (agc_loop == 0) {
		// open loop. Gain is fixed.
		return;
	} else if(((regv[39] & 0x0040) == 0) && ((FIO1PIN & 0x20000000) == 0)){
		// freeze IF AGC when in half-duplex mode and RX_TXN = '0'
		return;
	} else if (agc_loop == 1) {
		// closed internal IF AGC loop based on local IF1_POWER_DET measurement
		if (aux_adc[2] > 0x0E80) {
			// Near saturation detected. Reduce gain
			IF_gain_too_large_duration++;
			IF_gain_too_small_duration = 0; // reset the time elapsed since last peak detect > 0x80
			if (IF_gain_too_large_duration > 16) { // agc reaction time implied here
				IF_gain_too_large_duration = 0; // reset counter
				// decrement
				if (IF_gain > 0) {
					IF_gain--;
					setIFGain();
				} else {
					IF_gain = 0; // enforce range 0 -  4095
				}
			}
		} else {
			// slow pressure to increase the gain
			IF_gain_too_small_duration++; // increment the time elapsed since last saturation detect
			if (IF_gain_too_small_duration > 16) { // agc reaction time implied here
				// gain is consistently too small.
				IF_gain_too_large_duration = 0; // reset counter
				IF_gain_too_small_duration = 0; // reset the time elapsed since last saturation detect
				// increment
				if (IF_gain < 4095) {
					IF_gain++;
					setIFGain();
				} else {
					IF_gain = 4095; // enforce range 0 -  4095
				}
			}
		}
	} else if (agc_loop == 2) {
		// external AGC loop. Controlled by AGC_IN (40-pin connector, A6)
		// read ADC (converting the AGC_IN signal from the 40-pin connector to a 12-bit value)
		// flip gain control value (ComBlocks AGC rule: 0 is for maximum gain)
		IF_gain = 4095 - aux_adc[3];	// agc_in range 0 - 4095
		// set digital gain
		setIFGain();
	}
}

void RF_gain_increment(void) {
	if (RF_gain < 4095) {
		RF_gain++;
		setRFGain();
	} else {
		RF_gain = 4095; // enforce range 0 -  4095
	}
}

void RF_gain_decrement(void) {
	if (RF_gain > 0) {
		RF_gain--;
		setRFGain();
	} else {
		RF_gain = 0; // enforce range 0 -  4095
	}
}

//long pinDiodeAttenuatorCompensation(long gain){
//	// attempt at avoiding instabilities
//	// function of center frequency fc (global)
//	// input range 0 - 1023
//
//	long revised_gain;
//	revised_gain = gain;
//
//	if(fc < 200000000){
//		// below 200 MHz, avoid very large attenuations (Vc < 0.55V) which are destabilizing
//		if(gain < 129)
//			revised_gain = 129;
//	} else if (fc < 500000000){
//		// below 500 MHz, avoid very large attenuations (Vc < 0.5V) which are destabilizing
//		if(gain < 108)
//			revised_gain = 108;
//	} else if (fc > 2000000000){
//		// above 2 GHz, avoid too much gain (> 2.5V) because it can create instabilities
//		if(gain > 538)
//			revised_gain = 538;
//	}
//	return revised_gain;
//}

void ADC_init(void) {
	// ADC initialization
	PCONP |= 0x00001000; // power ADC subsystem
	// PCLKSEL0: by default peripheral clock is CCLK/4 = 30 MHz
	// A/D Control Register: inputs are AD0[2] and [5], ADC clock is 10 MHz, burst=1, operational,
	AD0CR = 0x00210224;
}

//void setLPF() {
//	// set the low-pass filter bandwidth
//	// REG4/REG5 stores the one-sided bandwidth in kHz
//	// Controls the CIC decimation filter and the two Half-Band filters
//	int bw, cic_r;
//
//	// read single-sided bandwidth in KHz
//	bw = regv[5] & 0x00FF;
//	bw = ((bw << 8) & 0x00FF00) | (regv[4] & 0x00FF);
//
//	// assuming a 100 MSamples/s sampling clock at the FPGA, let's calculate the
//	// decimation and the need to bypass
//	// we want to decimate from +/-fs/2 (at the ADC) to +/-4*bw (at the CIC filter output)
//	cic_r = (int) (50000 / (4 * bw));
//	// check valid range 1 - 16384
//	if (cic_r < 1)
//		cic_r = 1; // 0 is not an acceptable value
//	else if (cic_r > 16384)
//		cic_r = 16384;
//	regv[107] = (unsigned char) (cic_r & 0x00FF);
//	regv[108] = (unsigned char) ((cic_r >> 8) & 0x003F);
//
//	// bypass filters
//	if (bw > 12500) {
//		// bypass CIC + 1st HBF
//		regv[108] = (regv[108] & 0xC0) | 0x40;
//	} else if (bw > 6250) {
//		// bypass CIC
//		regv[108] = (regv[108] & 0xC0) | 0x80;
//	} else {
//		// enable all filters
//		regv[108] = (regv[108] & 0xC0) | 0xC0;
//	}
//
//	fpga_setByte(107, regv[107]);
//	fpga_setByte(108, regv[108]);
//
//}

//int AD7298_rw(int cw ){
//	// write control word cw, read data from ADC
//	// write control word = 0 when no control word to write
//	// returns 4 bit address + 12-bit sample
//	// usage: write FCD2 once to enable repetitive measurements on all channels except 7
//	int i, retval;
//	retval = 0;
//
//	PIN_AD7298_SCLK_H();
//	PIN_AD7298_CSB_L();
//	for(i = 0; i < 16; i++){
//		// read bit from DOUT
//		retval <<= 1;	// MSb first when reading
//		if(FIO2PIN & 0x080){	// P2[7]
//			retval |= 0x00001;
//		} else {
//			retval &= ~0x00001;
//		}
//		// write bit to DIN
//		if(cw & 0x08000){
//			PIN_AD7298_DIN_H();
//		} else {
//			PIN_AD7298_DIN_L();
//		}
//		cw <<= 1;	// MSb first when writing
//		PIN_AD7298_SCLK_L();
//		PIN_AD7298_SCLK_H();
//
//	}
//	PIN_AD7298_CSB_H();
//	return retval;
//}

int AD7298_rw(int cw ){
	// write control word cw, read data from ADC
	// write control word = 0 when no control word to write
	// returns 4 bit address + 12-bit sample
	// usage: write FCD2 once to enable repetitive measurements on all channels except 7
	// Re-wrote to shave 3us
	int retval;
	retval = 0;

	PIN_AD7298_SCLK_H();
	PIN_AD7298_CSB_L();
	// bit 15
	if(FIO2PIN & 0x080)	// P2[7]
		retval |= 0x08000;
	// write bit to DIN
	if(cw & 0x08000){
		PIN_AD7298_DIN_H();
	} else {
		PIN_AD7298_DIN_L();
	}
	PIN_AD7298_SCLK_L();
	PIN_AD7298_SCLK_H();
	// bit 14
	if(FIO2PIN & 0x080)	// P2[7]
		retval |= 0x04000;
	// write bit to DIN
	if(cw & 0x04000){
		PIN_AD7298_DIN_H();
	} else {
		PIN_AD7298_DIN_L();
	}
	PIN_AD7298_SCLK_L();
	PIN_AD7298_SCLK_H();
	// bit 13
	if(FIO2PIN & 0x080)	// P2[7]
		retval |= 0x02000;
	// write bit to DIN
	if(cw & 0x02000){
		PIN_AD7298_DIN_H();
	} else {
		PIN_AD7298_DIN_L();
	}
	PIN_AD7298_SCLK_L();
	PIN_AD7298_SCLK_H();
	// bit 12
	if(FIO2PIN & 0x080)	// P2[7]
		retval |= 0x01000;
	// write bit to DIN
	if(cw & 0x01000){
		PIN_AD7298_DIN_H();
	} else {
		PIN_AD7298_DIN_L();
	}
	PIN_AD7298_SCLK_L();
	PIN_AD7298_SCLK_H();
	// bit 11
	if(FIO2PIN & 0x080)	// P2[7]
		retval |= 0x00800;
	// write bit to DIN
	if(cw & 0x00800){
		PIN_AD7298_DIN_H();
	} else {
		PIN_AD7298_DIN_L();
	}
	PIN_AD7298_SCLK_L();
	PIN_AD7298_SCLK_H();
	// bit 10
	if(FIO2PIN & 0x080)	// P2[7]
		retval |= 0x00400;
	// write bit to DIN
	if(cw & 0x00400){
		PIN_AD7298_DIN_H();
	} else {
		PIN_AD7298_DIN_L();
	}
	PIN_AD7298_SCLK_L();
	PIN_AD7298_SCLK_H();
	// bit 9
	if(FIO2PIN & 0x080)	// P2[7]
		retval |= 0x00200;
	// write bit to DIN
	if(cw & 0x00200){
		PIN_AD7298_DIN_H();
	} else {
		PIN_AD7298_DIN_L();
	}
	PIN_AD7298_SCLK_L();
	PIN_AD7298_SCLK_H();
	// bit 8
	if(FIO2PIN & 0x080)	// P2[7]
		retval |= 0x00100;
	// write bit to DIN
	if(cw & 0x00100){
		PIN_AD7298_DIN_H();
	} else {
		PIN_AD7298_DIN_L();
	}
	PIN_AD7298_SCLK_L();
	PIN_AD7298_SCLK_H();
	// bit 7
	if(FIO2PIN & 0x080)	// P2[7]
		retval |= 0x00080;
	// write bit to DIN
	if(cw & 0x00080){
		PIN_AD7298_DIN_H();
	} else {
		PIN_AD7298_DIN_L();
	}
	PIN_AD7298_SCLK_L();
	PIN_AD7298_SCLK_H();
	// bit 6
	if(FIO2PIN & 0x080)	// P2[7]
		retval |= 0x00040;
	// write bit to DIN
	if(cw & 0x00040){
		PIN_AD7298_DIN_H();
	} else {
		PIN_AD7298_DIN_L();
	}
	PIN_AD7298_SCLK_L();
	PIN_AD7298_SCLK_H();
	// bit 5
	if(FIO2PIN & 0x080)	// P2[7]
		retval |= 0x00020;
	// write bit to DIN
	if(cw & 0x00020){
		PIN_AD7298_DIN_H();
	} else {
		PIN_AD7298_DIN_L();
	}
	PIN_AD7298_SCLK_L();
	PIN_AD7298_SCLK_H();
	// bit 4
	if(FIO2PIN & 0x080)	// P2[7]
		retval |= 0x00010;
	// write bit to DIN
	if(cw & 0x00010){
		PIN_AD7298_DIN_H();
	} else {
		PIN_AD7298_DIN_L();
	}
	PIN_AD7298_SCLK_L();
	PIN_AD7298_SCLK_H();
	// bit 3
	if(FIO2PIN & 0x080)	// P2[7]
		retval |= 0x00008;
	// write bit to DIN
	if(cw & 0x00008){
		PIN_AD7298_DIN_H();
	} else {
		PIN_AD7298_DIN_L();
	}
	PIN_AD7298_SCLK_L();
	PIN_AD7298_SCLK_H();
	// bit 2
	if(FIO2PIN & 0x080)	// P2[7]
		retval |= 0x00004;
	// write bit to DIN
	if(cw & 0x00004){
		PIN_AD7298_DIN_H();
	} else {
		PIN_AD7298_DIN_L();
	}
	PIN_AD7298_SCLK_L();
	PIN_AD7298_SCLK_H();
	// bit 1
	if(FIO2PIN & 0x080)	// P2[7]
		retval |= 0x00002;
	// write bit to DIN
	if(cw & 0x00002){
		PIN_AD7298_DIN_H();
	} else {
		PIN_AD7298_DIN_L();
	}
	PIN_AD7298_SCLK_L();
	PIN_AD7298_SCLK_H();
	// bit 0
	if(FIO2PIN & 0x080)	// P2[7]
		retval |= 0x00001;
	// write bit to DIN
	if(cw & 0x00001){
		PIN_AD7298_DIN_H();
	} else {
		PIN_AD7298_DIN_L();
	}
	PIN_AD7298_SCLK_L();
	PIN_AD7298_SCLK_H();


	PIN_AD7298_CSB_H();
	return retval;
}


void AD7298_poll(void){
	// poll auxiliary ADC and save 12-bit measurement for each of the (up to) 8 enabled channels.
	int c, channel;

	c = AD7298_rw(0);
	channel = (c >> 12) & 0x07;	// ADC channel number
	aux_adc[channel] = c & 0x0FFF;
}

void AD5621_1(int dac_val){
	// 12-bit auxiliary DAC (faster than internal DAC and PWMs)
	int i;
	dac_val = (dac_val << 2) & 0x3FFC;	// prepend 2 'normal operation' bits. append 2 x bits.

	PIN_AD5621_1_SYNCB_H();
	PIN_AD5621_1_SCLK_H();
	PIN_AD5621_1_SDIN_L();

	PIN_AD5621_1_SYNCB_L();
	for (i = 0; i < 16; i++) {
		if(dac_val & 0x8000){
			PIN_AD5621_1_SDIN_H();
		} else {
			PIN_AD5621_1_SDIN_L();
		}
		PIN_AD5621_1_SCLK_L();
		dac_val <<= 1;
		PIN_AD5621_1_SCLK_H();
	}
	PIN_AD5621_1_SYNCB_H();
}

void AD5621_2(int dac_val){
	// 12-bit auxiliary DAC (faster than internal DAC and PWMs)
	int i;
	dac_val = (dac_val << 2) & 0x3FFC;	// prepend 2 'normal operation' bits. append 2 x bits.

	PIN_AD5621_2_SYNCB_H();
	PIN_AD5621_2_SCLK_H();
	PIN_AD5621_2_SDIN_L();

	PIN_AD5621_2_SYNCB_L();
	for (i = 0; i < 16; i++) {
		if(dac_val & 0x8000){
			PIN_AD5621_2_SDIN_H();
		} else {
			PIN_AD5621_2_SDIN_L();
		}
		PIN_AD5621_2_SCLK_L();
		dac_val <<= 1;
		PIN_AD5621_2_SCLK_H();
	}
	PIN_AD5621_2_SYNCB_H();
}

