/*-------------------------------------------------------------
 *  Copyright (c) 2018, MSS ComBlock. All rights reserved.
 *
 * 	Filename: i2c.c
 *  Last modified: 1/18/18
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 0
 *  Inheritance: N/A
 *  Description: I2C configurations.
 *  See https://sites.google.com/site/johnkneenmicrocontrollers/18b-i2c/i2c_lpc1768#linkpwr
 * -------------------------------------------------------------*/
#include "main.h"
#include "i2c.h"
#include "utils.h"
#include "vectors.h"
#include "lpc_types.h"

// global
//tbd

void i2c2_init(int enable) {
	// configure I2C2
	if(!enable){
		// Power Control for Peripherals register
		PCONP &= ~0x04000000;
		return;
	} else {
		// Power Control for Peripherals register
		PCONP |= 0x04000000;
		// Peripheral Clock Selection CCLK (here 120/4 MHz)
		PCLKSEL1 &= ~0x00300000;
		// I2C2 pins P0[10,11]
		PINSEL0 = (PINSEL0 & ~0x00F00000) | 0x00A00000;
		// P0[10,11] as outputs
		FIO0DIR = (FIO0DIR & ~0x00000C00) | 0x00000C00;
		// no PU/PD
		PINMODE0 = (PINMODE0 & ~0x00F00000) | 0x00A00000;
		// open drain
		PINMODE_OD0 = (PINMODE_OD0 & ~0x00000C00) | 0x00000C00;
		// 400 KHz bit rate (see table 394)
		I22SCLH = 37;
		I22SCLL = 38;
//		// 100 KHz bit rate (see table 394)
//		I22SCLH = 150;
//		I22SCLL = 150;
		// enable master transmitter mode
		I22CONSET |= 0x40;	// set I2EN
		return;
	}
}

int i2c2_tx(int addr, int regaddr, int regdata){
	// send one byte data to specified address
	// function is blocking (will not return until transfer is complete)
	// return 0 if successful

	// enter master transmit mode: set the STA bit
	I22CONSET |= 0x20;
	// wait until SI interrupt
	while(!(I22CONSET & 0x08)){;}
	if(I22STAT != 0x08){
		// did not receive the expected "A START condition has been transmitted."
		return 1;
	}
	// load slave address, add write bit
	I22DAT = addr & 0x00FE;
	// clear start bit
	I22CONCLR = 0x20;	// clear STA
	I22CONCLR = 0x08;	// clear SI
	// wait until SI interrupt
	while(!(I22CONSET & 0x08)){;} //wait until change in status
	if(I22STAT != 0x18){
		// did not receive the expected ACK
		return 2;
	}
	// load 1st data byte (MAX2121 register address)
	I22DAT = regaddr & 0x00FF;
	I22CONCLR = 0x08;	// clear SI
	// wait until SI interrupt
	while(!(I22CONSET & 0x08)){;} //wait until change in status
	if(I22STAT != 0x28){
		// did not receive the expected ACK
		return 3;
	}
	// load 2nd data byte (MAX2121 register value)
	I22DAT = regdata & 0x00FF;
	I22CONCLR = 0x08;	// clear SI
	// wait until SI interrupt
	while(!(I22CONSET & 0x08)){;} //wait until change in status
	if(I22STAT != 0x28){
		// did not receive the expected ACK
		return 4;
	}
	// STOP
	I22CONSET |= 0x10;
	I22CONCLR = 0x08;	// clear SI
	// wait until hardware stops
	while(!(I22CONSET & 0x10)){;}


	return 0;
}

int i2c2_rx(int addr, int regaddr){
	// read one byte data from specified address
	// function is blocking (will not return until transfer is complete)
	int retdata;

	// enter master receive mode: set the STA bit
	I22CONSET |= 0x20;
	// wait until SI interrupt
	while(!(I22CONSET & 0x08)){;}
	if(I22STAT != 0x08){
		// did not receive the expected "A START condition has been transmitted."
		return 0x100;
	}
	// load slave address, add write bit
	I22DAT = addr & 0x00FE;
	// clear start bit
	I22CONCLR = 0x20;	// clear STA
	I22CONCLR = 0x08;	// clear SI
	// wait until SI interrupt
	while(!(I22CONSET & 0x08)){;} //wait until change in status
	if(I22STAT != 0x18){
		// did not receive the expected ACK
		return 0x200;
	}
	// load 1st data byte (MAX2121 register address)
	I22DAT = regaddr & 0x00FF;
	I22CONCLR = 0x08;	// clear SI
	// wait until SI interrupt
	while(!(I22CONSET & 0x08)){;} //wait until change in status
	if(I22STAT != 0x28){
		// did not receive the expected ACK
		return 0x300;
	}
	// set the STA bit
	I22CONSET |= 0x20;
	I22CONCLR = 0x08;	// clear SI
	// wait until SI interrupt
	while(!(I22CONSET & 0x08)){;}
	if(I22STAT != 0x10){
		// did not receive the expected "A repeat START condition has been transmitted."
		return 0x400;
	}
	// load slave address, add read bit
	I22DAT = (addr | 0x01) & 0x00FF;
	// clear start bit
	I22CONCLR = 0x20;	// clear STA
	I22CONCLR = 0x08;	// clear SI
	// wait until SI interrupt
	while(!(I22CONSET & 0x08)){;} //wait until change in status
	if(I22STAT != 0x40){
		// did not receive the expected ACK
		return 0x500;
	}
	// no ACK, last byte
	I22CONCLR = 0x04;	// no ACK
	I22CONCLR = 0x08;	// clear SI
	// wait until SI interrupt
	while(!(I22CONSET & 0x08)){;} //wait until change in status
	retdata = I22DAT & 0x00FF;
	// STOP
	I22CONSET |= 0x10;
	I22CONCLR = 0x08;	// clear SI
	// wait until hardware stops
	while(!(I22CONSET & 0x10)){;}

	return retdata;
}
