/*-------------------------------------------------------------
 *  Copyright (c) 2010-2017, MSS ComBlock. All rights reserved.
 *
 * 	Filename: main.c
 *  Last modified: 2/23/17
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 2.2
 *  Inheritance: CB3011
 *  Description: main program. Includes main loop for handling commands on a polling basis.
 *
 *  Dependencies/Pre-requisite: N/A
 *
 *  Usage:
 *  Stores 8 frequencies expressed in Hz in
 *  frequency index 0: REG3/2/1/0
 *  frequency index 1: REG10/9/8/7
 *  frequency index 2: REG14/13/12/11
 *  frequency index 3: +4 bytes
 *  etc.
 *  Tx frequency index is in REG6[7:4]
 *  Rx frequency index is in REG6[3:0]
 *  Rx LNA initial gain: REG40 (LSB), REG41(3:0) MSbs
 *  Rx RF initial gain: REG4 (LSB), REG5(3:0) MSbs
 *  Rx IF initial gain: REG35 (LSB), REG36(3:0) MSbs
 *  Tx gain: REG37 (LSB), REG38(1:0) MSbs
 *  AGC control: REG39
 *    bits 0  LNA AGC open loop, fixed (0), local loop (1)
 *    bits 1  RF AGC open loop, fixed (0), local loop (1)
 *    bits 3:2  IF AGC open loop, fixed by control registers (00), local loop (01), external loop (10)
 *    bits 5:4  tx gain control, fixed by control register(00), external analog TX_GAIN_CTRL1 (01)
 *
 *  Control changes are enacted upon (re-)writing to the last control register (REG38)
 *
 *  Status registers:
 *  sreg0: power good 1-8
 *  sreg1: power good 9
 *  sreg2: rf synthesizers lock
 *  	bit 0 = rx
 *  	bit 1 = tx
 *  sreg3: RF_POWER_DET1  RSSI[7:0]
 *  sreg4: RF_POWER_DET1  RSSI[11:8]
 *  sreg5: RF_POWER_DET2  power detection from RF mixer[7:0]
 *  sreg6: RF_POWER_DET2  power detection from RF mixer[11:8]
 *  sreg7: IF1_POWER_DET IF1 power detection [7:0]
 *  sreg8: IF1_POWER_DET IF1 power detection[11:8]
 *  sreg9: Board temperature sensor[7:0]
 *  sreg10: Board temperature sensor[11:8]
 *  sreg11: LNA gain[7:0]
 *  sreg12: LNA gain[11:8]
 *  sreg13: RF gain[7:0]
 *  sreg14: RF gain[11:8]
 *  sreg15: RF gain[7:0]
 *  sreg16: RF gain[11:8]
 *
 * Rev 1.1 1/8/15 AZ
 * Added IF_CENTER_FREQ as a #define (select 140MHz by default, or 70 MHz upon request)
 *
 * Rev 2 2 8/10/15 AZ
 * Added TX_EN poll to squelch the RF output when not transmitting (no LO leakage)
 * 3 versions 70 MHz, 140 MHz, 120 MHz rx IF (see main.h)
 *
 *  Rev 2.1 12/12/15 AZ
 *  Added usb connection/disconnection in main, depending on the presence of the USB VBUS
 *
 *  Rev 2.2 2/23/17 AZ
 *  Phase noise improvement (removed ref clk doubling, same as com-4009)
 -------------------------------------------------------------*/
#define __MAIN_C__

#include "main.h"
#include "uart.h"
#include "utils.h"
#include "iap.h"
#include "rf.h"

/*=========================================================================*/
/*  DEFINE: All Structures and Common Constants                            */
/*=========================================================================*/
volatile unsigned long SysTickCnt; /* SysTick Counter                    */
unsigned long cpuFreq; // CPU frequency in Hz.
volatile unsigned long timer[NTIMERS];
// timers, expressed in multiple of the system tick (10ms). 0 means expired.
// volatile because the Systick ISR will modify the values
// timer[0] = flash timer
// timer[1] = forced delay to save the config into flash
// timer[2] = inactivity timer
volatile string_struct rxLine; // rx command line
volatile string_struct txLine; // tx command line
// separate rx and tx lines (unlike previous Atmel code). Allows concurrent tx and rx and interrupts.
char address[4]; // ComBlock address, assigned dynamically (string of 3 characters + \0 end of string)
long commandPortNo; // port through which this board receives commands: 1 = top (USB ARM), 2 = right, 3 = bottom (USB FPGA), 4 = left (ComBlock rule)
unsigned char forwardMask; // Mask output ports. bit0 = top port1, bit1 = right port 2, etc. '1' means the port is blocked, '0' means the port is enabled for output
struct config_hw configHw; // hardware-related configuration.
struct config_sw configSw; // software/FPGA-related configuration volatile
char regv[NREGS]; // control registers, volatile copy of configSw.reg[]
int active_sw_config; // current active configuration (generally the same as default_sw_config, but it could be temporary different until the new FPGA config is loaded)
int usb_softconnect;	// 0 for no vbus/no connect, 1 for vbus/connected
unsigned int LNA_gain;	// receiver LNA gain (1 RF attenuator between LNA and first gain block)
unsigned int RF_gain;	// receiver RF gain (2 RF attenuators)
unsigned int IF_gain;	// receiver IF gain (1 attenuator)
long RF_gain_too_small_duration; // below normal range samples timer
long RF_gain_too_large_duration; // above normal range samples timer
long IF_gain_too_small_duration; // below normal range samples timer
long IF_gain_too_large_duration; // above normal range samples timer
long fc;	// estimated receiver center frequency (needed to compute attenuators response)
unsigned long agc_in;	// gain control value from the 40-pin connector, low-pass filtered = AD0[5]
unsigned int aux_adc[8];	// 12-bit auxiliary ADC measurements

/***************************************************************************/
/*  main                                                                   */
/***************************************************************************/
int main(void) {
	// local variables
	unsigned char a;
	int ssp1_rx_word;
	unsigned int i, c, reg_no, s78_decode, s789_decode,
			nreg_no, start_reg_no;
	int sn[4];
	struct config_sw cs; // work area for a software configuration (don't store anything of value here)
	char sreg1, sreg4, sreg6, sreg8, sreg10, sreg12, sreg14, sreg16;	// temporary storage for multi-byte monitoring values

	// initial ARM processor configuration (clock frequency, system tick, etc)
	initConfig();

	// initialize UARTs 0,3.
	rst: uart_init();

	// initialize SSP1
	ssp1_init(1);

	// When the internal flash sector CONFIG_SECTOR is blank (all FFs), initialize it to 'all zeros'.
	initialize_configs();
	// load last-saved configuration from non-volatile (flash) memory
	read_hw_config(&configHw);

	// Only one personality.
	active_sw_config = 0;

	// Semantics: we make a difference between the active_sw_config (currently loaded and running) and
	// default_sw_config which can be changed by the user prior to FPGA reboot. N/A here (one configuration)
	read_sw_config(active_sw_config, &configSw);
	memcpy(&regv, configSw.reg, sizeof(regv)); // initialize the temporary control registers with the previously saved ones.

	// Tx gain control
	// initialize internal 10-bit DAC, minimize transmit level temporarily while FPGA is not configured
	DAC_init();
	DAC(0);	// minimize TX_GAIN_CTRL2
	PIN_RFOUT_ENB_H();	// disable the RF modulator output (active low)

	// 2 second delay at power up to give the FPGA time for initial configuration.
	// Otherwise the power amplifier will be on blasting noise (and possibly destroying any sensitive
	// test equipment)
	delay_ms(2000);

	// ComBlock address at power up is "000"
	snprintf(&address[0],sizeof(address), "000");

	// RF frequency synthesizer (ADF4351) initialization
	if((regv[42] & 0x0F) == 0) {
		// 19.2 MHz internal (TCXO) frequency reference
		init_tuner_pll(1,3); // one time initialization
		init_tuner_pll(2,3); // one time initialization
		program_tuner_pll(1,3); // Program the default frequency
		program_tuner_pll(2,3); // Program the default frequency
	} else {
		// 10 MHz external frequency reference
		init_tuner_pll(1,4); // one time initialization
		init_tuner_pll(2,4); // one time initialization
		program_tuner_pll(1,4); // Program the default frequency
		program_tuner_pll(2,4); // Program the default frequency
	}

	// start auxiliary ADC
	AD7298_rw(0xFFC2);	// enable all channel 0,1,2,3,4,5,6,7 , repeat, no temperature

	// receiver gain (LNA). Range 0 - 4095
	PWM_init();
	LNA_gain = (int)(regv[41]) & 0x0000000F;
	LNA_gain <<= 8;
	LNA_gain |= (int)(regv[40]) & 0x000000FF;
	setLNAgain();	// set initial LNA gain

	// receiver RF gain. Initial value before RF AGC takes over. range 0 - 4095
	RF_gain = (int)(regv[5]) & 0x0000000F;
	RF_gain <<= 8;
	RF_gain |= (int)(regv[4]) & 0x000000FF;
	setRFGain();	// set initial RF gain

	// receiver IF gain. Initial value before baseband AGC takes over. range 0 - 4095
	IF_gain = (int)(regv[36]) & 0x0000000F;
	IF_gain <<= 8;
	IF_gain |= (int)(regv[35]) & 0x000000FF;
	setIFGain();	// set initial baseband gain

	// Tx gain control
	TX_EN_poll();
	if(regv[39] & 0x0040){
		// full duplex operation. Ignore RX_TX#
		// disable GPIO interrupt for RX_TX# rising/falling
		IO0_INT_EN_R = 0x00000000;	// p0[25]  RX_TX#, rising edge
		IO0_INT_EN_F = 0x00000000;	// p0[25]  RX_TX#, falling edge
		PIN_SYNTH1_ENABLE_H();
		PIN_SYNTH2_ENABLE_H();
		PIN_RFOUT_ENB_L();	// enable the RF modulator output (active low)
		setTxGain();
	} else {
		// half-duplex
		// enable GPIO interrupt for RX_TX# rising/falling
		IO0_INT_EN_R = 0x02000000;	// p0[25]  RX_TX#, rising edge
		IO0_INT_EN_F = 0x02000000;	// p0[25]  RX_TX#, falling edge
		RX_TXN_poll();
	}

	// main loop -------------------------------------------------------------------------------
	while (1) {
		// USB softconnect management based on USB VBUS presence
		USB_connect_ctrl();

		// is it time to save the current active configuration?
		// timer[1] limits how frequently the configuration can be saved,
		// timer[2] indicates persistent inactivity.
		if (((configHw.save_request_pending) || (configSw.save_request_pending))
				&& (timer[2] == 0) && (timer[1] == 0)) {
			save_configuration(-1, &cs); // save the hardware + active software configurations. no alternate configuration.
		}

		// service UARTs (polling)
		uart_rxpoll();
		uart_txpoll();

		// poll the auxiliary ADC (8 channels enabled)
		AD7298_poll();

		// rx LNA gain control.
		LNA_AGC_poll();
		// rx RF gain control.
		RF_AGC_poll();
		// rx IF gain control
		IF_AGC_poll();
		// check TX_EN
		TX_EN_poll();

		if(regv[39] & 0x0040){
			// full duplex operation. Ignore RX_TX#
			tx_gain_poll();
		} else if (!(FIO0PIN & 0x02000000)){
			// half-duplex while in tx mode (RX_TX# = '0')
			tx_gain_poll();
		}

		// monitor board status
		if (!synth_check_lock(1)){
			PIN_RED_LED_H();
		} else if (!synth_check_lock(2)){
			PIN_RED_LED_H();
		} else {
			// all OK
			PIN_RED_LED_L();
		}

		// poll SSP/SPI interface for receive signals
		ssp1_rx_word = ssp1_rxFIFO_poll();
		if(ssp1_rx_word >= 0){
			// received 16-bit word from SPI master.
			reg_no = (ssp1_rx_word >> 8) & 0x007F;	// 7-bit control/status register address
			if(ssp1_rx_word & 0x00008000){
				// write command (equivalent to the set register SRG command over serial/USB)
				a = ssp1_rx_word & 0x000000FF;
				// all commands over SSP are stored in non-volatile registers
				if (reg_no < sizeof(configSw.reg)) {
					regv[reg_no] = a;
				}
				// configure (same code as in SRG section below)
				if (reg_no == 6) {
					// RF synthesizers reprogrammed upon writing to control REG6
					if((regv[42] & 0x0F) == 0) {
						// 19.2 MHz internal (TCXO) frequency reference
						program_tuner_pll(1,3); // Program the default frequency
						program_tuner_pll(2,3); // Program the default frequency
					} else {
						// 10 MHz external frequency reference
						program_tuner_pll(1,4); // Program the default frequency
						program_tuner_pll(2,4); // Program the default frequency
					}
				} else if (reg_no == 5) {
					// receiver RF gain. Initial value before RF AGC takes over. range 0 - 4095
					RF_gain = (int)(regv[5]) & 0x0000000F;
					RF_gain <<= 8;
					RF_gain |= (int)(regv[4]) & 0x000000FF;
					setRFGain();	// set initial RF gain
				} else if (reg_no == 36) {
					// receiver IF gain. Initial value before baseband AGC takes over. range 0 - 4095
					IF_gain = (int)(regv[36]) & 0x0000000F;
					IF_gain <<= 8;
					IF_gain |= (int)(regv[35]) & 0x000000FF;
					setIFGain();	// set initial baseband gain
				} else if (reg_no == 39){
					if(regv[39] & 0x0040){
						// full duplex operation. Ignore RX_TX#
						// disable GPIO interrupt for RX_TX# rising/falling
						IO0_INT_EN_R = 0x00000000;	// p0[25]  RX_TX#, rising edge
						IO0_INT_EN_F = 0x00000000;	// p0[25]  RX_TX#, falling edge
						PIN_SYNTH1_ENABLE_H();
						PIN_SYNTH2_ENABLE_H();
						PIN_RFOUT_ENB_L();	// enable the RF modulator output (active low)
						setTxGain();
					} else {
						// enable GPIO interrupt for RX_TX# rising/falling
						IO0_INT_EN_R = 0x02000000;	// p0[25]  RX_TX#, rising edge
						IO0_INT_EN_F = 0x02000000;	// p0[25]  RX_TX#, falling edge
						RX_TXN_poll();
					}
				} else if (reg_no == 41) {
					// receiver LNA gain (10-bit)
					LNA_gain = (int)(regv[41]) & 0x0000000F;
					LNA_gain <<= 8;
					LNA_gain |= (int)(regv[40]) & 0x000000FF;
					setLNAgain();	// set LNA gain
				}

			} else {
				// read command (equivalent to the get status register GSR command over serial/USB)
				// same code as in the GSR section below
				if (reg_no == 0) {
					// power good signals part 1
					c = power_check();
					sreg1 = (c >> 8) & 0x0FF;	// save for 2nd part query
					a = c & 0x0FF;
				} else	if (reg_no == 1) {
					// power good signals part 2
					a = sreg1;
				} else	if (reg_no == 2) {
					// RF synthesizer lock status
					if (synth_check_lock(1))
						a |= 0x01; // SREG2 bit 0
					if (synth_check_lock(2))
						a |= 0x02; // SREG2 bit 1

				} else	if (reg_no == 3) {
					// RSSI (RF_POWER_DET1), LSB
					a = aux_adc[0] & 0x00FF;
					sreg4 = (aux_adc[0] >> 8) & 0x00F;
				} else	if (reg_no == 4) {
					// RSSI (RF_POWER_DET1), MSbs
					a = sreg4;

				} else	if (reg_no == 5) {
					// RF power detect from the RF mixer (RF_POWER_DET2), LSB
					// save MSbs for later
					a = aux_adc[1] & 0x00FF;
					sreg6 = (aux_adc[1] >> 8) & 0x00F;
				} else	if (reg_no == 6) {
					// RF power detect from the RF mixer (RF_POWER_DET2), MSbs
					a = sreg6;

				} else	if (reg_no == 7) {
					// IF1 power detect (IF1_POWER_DET), LSB
					a = aux_adc[2] & 0x00FF;
					sreg8 = (aux_adc[2] >> 8) & 0x00F;
				} else	if (reg_no == 8) {
					// IF1 power detect (IF1_POWER_DET), MSbs
					a = sreg8;

				} else	if (reg_no == 9) {
					// Board temperature sensor, LSB
					a = aux_adc[6] & 0x00FF;
					sreg10 = (aux_adc[6] >> 8) & 0x00F;
				} else	if (reg_no == 10) {
					// Board temperature sensor, MSbs
					a = sreg10;

				} else	if (reg_no == 11) {
					// LNA gain (LSB)
					a = LNA_gain & 0x00FF;
					sreg12 = (LNA_gain >> 8) & 0x00F;
				} else	if (reg_no == 12) {
					// LNA gain, MSbs
					a = sreg12;
				} else	if (reg_no == 13) {
					// RF gain, LSB
					a = RF_gain & 0x00FF;
					sreg14 = (RF_gain >> 8) & 0x00F;
				} else	if (reg_no == 14) {
					// RF gain, MSbs
					a = sreg14;
				} else	if (reg_no == 15) {
					// IF gain, LSB
					a = IF_gain & 0x00FF;
					sreg16 = (IF_gain >> 8) & 0x00F;
				} else	if (reg_no == 16) {
					// IF gain, MSbs
					a = sreg16;
				}
				c = ((unsigned int) a) & 0x000000FF;
				c |= ssp1_rx_word & 0xFFFFFF00;
				ssp1_txFIFO_write(c);
			}
		}
		// did we receive any message?
		// get a complete message from @ to \n (non-blocking)
		if ((rxLine.msg_complete) && (rxLine.rptr == 0)) {
			// When done processing the received message in rxLine, move the rptr = wptr to allow subsequent messages to be received.

			// re-arm inactivity timer because of new message just received.
			// We need to know when there is inactivity so that we can devote 120ms to update the configuration in flash.
			timer[2] = TIMER2;

			/*
			 // echo message (debugging)
			 memcpy(txLine.buf, rxLine.buf, rxLine.wptr);
			 txLine.wptr = rxLine.wptr;
			 rxLine.rptr = rxLine.wptr;	// release rx message
			 sendLine(4);
			 continue;
			 */

			// Dispatch the message: local processing, ignore or forward?
			if (dispatchLine()) {
				// rx message is for us: address at the start of the rx message matches (exactly or 111 broadcast address)
				delay_ns(25000); // delay 25uS before responding (otherwise slower Atmel-based ComBlocks may not be ready for a reply)


				// Remember which port is used to communicate with ComBlock control center
				// Criteria: received complete command message destined to this particular ComBlock.
				commandPortNo = rxLine.portNo;

				// decode 2-digit decimal reg_no at position 7-8 in the received string. Used by most commands.
				s78_decode = rxLine.buf[7] - '0';
				s78_decode = s78_decode * 10 + rxLine.buf[8] - '0';

				// decode 3-digit decimal sector number at position 7-9 in the received string (common to many commands)
				s789_decode = s78_decode * 10 + rxLine.buf[9] - '0';


				// command handler
				if (!memcmp((const char *)&rxLine.buf[4], "SAC", 3)) {
					// SAC command - set address
					memcpy(&address[0], (const char *)&rxLine.buf[7], 3);

				} else if (!memcmp((const char *)&rxLine.buf[4], "MFW", 3)) {
					// MFW command - Message Forwarding
					// 0 = no forwarding, 1-4: forward to port 1-4
					// 9: forward to all ports (except of course the one where the message came from)
					if (rxLine.buf[7] == '0') { // stop all forwarding
						forwardMask = 0xFF;
					} else if (rxLine.buf[7] == '1') { // forward to top port
						forwardMask = 0xFE;
					} else if (rxLine.buf[7] == '2') { // forward to right port
						forwardMask = 0xFD;
					} else if (rxLine.buf[7] == '3') { // forward to bottom port
						forwardMask = 0xFB;
					} else if (rxLine.buf[7] == '4') { // forward to left port
						forwardMask = 0xF7;
					} else if (rxLine.buf[7] == '9') { // forward to all ports, but the command port itself
						if (commandPortNo == 1) {
							forwardMask = 0xF1;
						} else if (commandPortNo == 2) {
							forwardMask = 0xF2;
						} else if (commandPortNo == 3) {
							forwardMask = 0xF4;
						} else if (commandPortNo == 4) {
							forwardMask = 0xF8;
						}
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "PNG", 3)) {
					// new PNG ping command. Echoes command back
					memcpy((char *)&(txLine.buf[0]), (const char *)&(rxLine.buf[0]), rxLine.wptr);
					txLine.buf[1] = '9';
					txLine.buf[2] = '9';
					txLine.buf[3] = '9';
					txLine.wptr = rxLine.wptr;
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "GMI", 3)) {
					// GMI - Get Module Identification command
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999MID" ARM_FW_REV "1\r\n");
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo); // send response through the same port the query came in

				} else if (!memcmp((const char *)&rxLine.buf[4], "SDC", 3)) {
					// SDC command - set default configuration
					// decode selected Configuration number
					c = s78_decode - 1;
					// make sure configuration pointer is within allowed boundaries
					if ((c != configHw.default_sw_config) && (c < NSWCONFIGS)) {
						// the default config changed.
						configHw.default_sw_config = c;
						configHw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
						// immediate save (as this command is likely to be followed by an immediate reset)
						timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
						// Note: saving takes 120ms!! during which incoming messages are not processed
						save_configuration(-1, &cs);
						//NOTE: at this point, active and default configuration numbers are different.
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "GDC", 3)) {
					// GDC command - Get Default Configuration command
					// prepare DCF response
					c = configHw.default_sw_config + 1; // display configuration number starting at 1 (more natural for the user)
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999DCF%02X\r\n", c);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "SMI", 3)) {
					// SMI command - Set Module Identification

					// first check the length of the message.
					if ((rxLine.buf[12] == '\r') || (rxLine.buf[12] == '\n')) {
						// message is of the form @xxxSMI + MID + 1character option
						// thus 12 or 13 characters. use active configuration.
						// verify that the selected configuration is not protected unless
						// it is the first time this configuration is defined (all FF's)
						if ((!configSw.protection) || (configSw.mid[0] == 0xFF)) {
							memcpy(&configSw.mid[0], (const char *)&rxLine.buf[7], 5); // store MID + 1character option
							if (rxLine.buf[11] == '\r') // short message @xxxSMIYYYY\r\n
								configSw.mid[4] = ' '; // option cannot be '\r'
							configSw.mid[5] = 0; // don't know about FPGA firmware revision here
							configSw.protection = 0; // unprotected if first-time definition
							configSw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
							timer[1] = TIMER1SLOW;
							// re-arm timer[1]: wait before saving the comblock configuration
							//into flash (to consolidate multiple successive changes into one sector programming operation)
						}
					} else if ((rxLine.buf[14] == '\r') || (rxLine.buf[14]
							== '\n')) {
						// message is of the form @xxxSMI+configuration number + MID + 1character option
						// thus 15 characters
						c = s78_decode - 1; // array indices start at 0, whereas configurations start at 1
						// make sure configuration pointer does not exceed array boundaries
						if (c < NSWCONFIGS) {
							read_sw_config(c, &cs); // read the software configuration of interest from flash
							// verify that the selected configuration is not protected unless it is the first time this configuration is defined (all FF's)
							if ((!cs.protection) || (cs.mid[0] == 0xFF)) {
								memcpy(&cs.mid[0], (const char *)&rxLine.buf[9], 5); // store MID + 1character option
								cs.mid[5] = 0; // don't know about FPGA firmware revision here
								cs.protection = 0; // unprotected if first-time definition
								// immediate save this non-active configuration
								timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
								// Note: saving takes 120ms!! during which incoming messages are not processed
								save_configuration(c, &cs);
							}
						}
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "SAI", 3)) {
					// SAI command - Set assembly identification
					// For security, allow only if not already programmed
					// first check if assembly id has already been programmed
					a = 0;
					/* temporarily disabled
					 for(i = 0; i < 4; i++){
					 if(aid[i] != 0)
					 a = 1;
					 }
					 */
					if (a == 0) {
						// if aid is undefined (all 0s), allow programming
						memcpy(&configHw.aid[0], (const char *)&rxLine.buf[7], 6);
						configHw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
						timer[1] = TIMER1SLOW;
						// re-arm timer[1]: wait before saving the comblock configuration
						//into flash (to consolidate multiple successive changes into one sector programming operation)
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "SAS", 3)) {
					// SAS - Set Assembly Serial number,
					// For security, allow only if eeprom is not already programmed (i.e. 0000..)
					// first check if serial number has already been programmed
					a = 0;
					for (i = 0; i < sizeof(configHw.asn); i++) {
						if (configHw.asn[i] != 0)
							a = 1;
					}
					if (a == 0) {
						// if assembly serial number is null, allow programming
						memcpy(&configHw.asn[0], (const char *)&rxLine.buf[7],
								sizeof(configHw.asn));
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "GAI", 3)) {
					// GAI command - Get Assembly ID
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999AID      \r\n"); // workaround problem with sprintf.
					memcpy((char *)&txLine.buf[7], (const char *)configHw.aid, 6);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "GSN", 3)) {
					// GSN command - Get Serial Number
					// read unique LPC17xx device serial number
					read_sn(sn);
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999MSN%10u %10u %10u %10u\r\n",
							sn[0], sn[1], sn[2], sn[3]); // unsigned decimal
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "GAS", 3)) {
					// GAS - Get Assembly Serial number command
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999ASN%10s\r\n", configHw.asn);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if ((!memcmp((const char *)&rxLine.buf[4], "SRG", 3)) || (!memcmp(
						(const char *)&rxLine.buf[4], "SRT", 3))) {
					// SRG - Set Register (persistent/non-volatile storage)
					// SRT - Set Register Temporary (i.e. volatile, lost when powering off)

					// compute register address. Address comprises at least two digits.
					c = s78_decode;
					// check whether destination register is 2 digit or 3 digit long
					if ((rxLine.buf[11] != '\r') && (rxLine.buf[11] != '\n')) {
						// three-digit register number
						c = c * 10 + rxLine.buf[9] - '0';
						// compute register value. Value starts at char 10
						a = hex2bin((char *)&rxLine.buf[10]);
					} else {
						// two-digit register number
						// compute register value. Value starts at char 9
						a = hex2bin((char *)&rxLine.buf[9]);
					}
					reg_no = c;

					// update the configuration register
					if (reg_no < sizeof(configSw.reg)) {
						regv[reg_no] = a;
						// if SRG, also save in flash
						if (rxLine.buf[6] == 'G') {
							configSw.reg[reg_no] = a;
							configSw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
							timer[1] = TIMER1SLOW;
							// re-arm timer[1]: wait before saving the comblock configuration
							//into flash (to consolidate multiple successive changes into one sector programming operation)
						}
					}
					if (reg_no == 6) {
						// RF synthesizers reprogrammed upon writing to control REG6
						if((regv[42] & 0x0F) == 0) {
							// 19.2 MHz internal (TCXO) frequency reference
							program_tuner_pll(1,3); // Program the default frequency
							program_tuner_pll(2,3); // Program the default frequency
						} else {
							// 10 MHz external frequency reference
							program_tuner_pll(1,4); // Program the default frequency
							program_tuner_pll(2,4); // Program the default frequency
						}
					} else if (reg_no == 5) {
						// receiver RF gain. Initial value before RF AGC takes over. range 0 - 4095
						RF_gain = (int)(regv[5]) & 0x0000000F;
						RF_gain <<= 8;
						RF_gain |= (int)(regv[4]) & 0x000000FF;
						setRFGain();	// set initial RF gain
					} else if (reg_no == 36) {
						// receiver IF gain. Initial value before baseband AGC takes over. range 0 - 4095
						IF_gain = (int)(regv[36]) & 0x0000000F;
						IF_gain <<= 8;
						IF_gain |= (int)(regv[35]) & 0x000000FF;
						setIFGain();	// set initial baseband gain
					} else if (reg_no == 39){
						if(regv[39] & 0x0040){
							// full duplex operation. Ignore RX_TX#
							// disable GPIO interrupt for RX_TX# rising/falling
							IO0_INT_EN_R = 0x00000000;	// p0[25]  RX_TX#, rising edge
							IO0_INT_EN_F = 0x00000000;	// p0[25]  RX_TX#, falling edge
							PIN_SYNTH1_ENABLE_H();
							PIN_SYNTH2_ENABLE_H();
							PIN_RFOUT_ENB_L();	// enable the RF modulator output (active low)
							setTxGain();
						} else {
							// enable GPIO interrupt for RX_TX# rising/falling
							IO0_INT_EN_R = 0x02000000;	// p0[25]  RX_TX#, rising edge
							IO0_INT_EN_F = 0x02000000;	// p0[25]  RX_TX#, falling edge
							RX_TXN_poll();
						}
					} else if (reg_no == 41) {
						// receiver LNA gain (10-bit)
						LNA_gain = (int)(regv[41]) & 0x0000000F;
						LNA_gain <<= 8;
						LNA_gain |= (int)(regv[40]) & 0x000000FF;
						setLNAgain();	// set LNA gain
					}


				} else if (!memcmp((const char *)&rxLine.buf[4], "GSR", 3)) {
					// GSR command - Get Status Register
					// check whether 1,2 or 3 digits
					c = rxLine.buf[7] - '0';
					if ((rxLine.buf[8] != '\r') && (rxLine.buf[8] != '\n')) {
						// 2 or 3 digit register number
						c = c * 10 + rxLine.buf[8] - '0';
					}
					if ((rxLine.buf[9] != '\r') && (rxLine.buf[9] != '\n')) {
						// 3 digit register number
						c = c * 10 + rxLine.buf[9] - '0';
					}
					reg_no = c;

					a = 0;
					if (reg_no == 0) {
						// power good signals part 1
						c = power_check();
						sreg1 = (c >> 8) & 0x0FF;	// save for 2nd part query
						a = c & 0x0FF;
					} else	if (reg_no == 1) {
						// power good signals part 2
						a = sreg1;
					} else	if (reg_no == 2) {
						// RF synthesizer lock status
						if (synth_check_lock(1))
							a |= 0x01; // SREG2 bit 0
						if (synth_check_lock(2))
							a |= 0x02; // SREG2 bit 1

					} else	if (reg_no == 3) {
						// RSSI (RF_POWER_DET1), LSB
						a = aux_adc[0] & 0x00FF;
						sreg4 = (aux_adc[0] >> 8) & 0x00F;
					} else	if (reg_no == 4) {
						// RSSI (RF_POWER_DET1), MSbs
						a = sreg4;

					} else	if (reg_no == 5) {
						// RF power detect from the RF mixer (RF_POWER_DET2), LSB
						// save MSbs for later
						a = aux_adc[1] & 0x00FF;
						sreg6 = (aux_adc[1] >> 8) & 0x00F;
					} else	if (reg_no == 6) {
						// RF power detect from the RF mixer (RF_POWER_DET2), MSbs
						a = sreg6;

					} else	if (reg_no == 7) {
						// IF1 power detect (IF1_POWER_DET), LSB
						a = aux_adc[2] & 0x00FF;
						sreg8 = (aux_adc[2] >> 8) & 0x00F;
					} else	if (reg_no == 8) {
						// IF1 power detect (IF1_POWER_DET), MSbs
						a = sreg8;

					} else	if (reg_no == 9) {
						// Board temperature sensor, LSB
						a = aux_adc[4] & 0x00FF;
						sreg10 = (aux_adc[4] >> 8) & 0x00F;
					} else	if (reg_no == 10) {
						// Board temperature sensor, MSbs
						a = sreg10;

					} else	if (reg_no == 11) {
						// LNA gain (LSB)
						a = LNA_gain & 0x00FF;
						sreg12 = (LNA_gain >> 8) & 0x00F;
					} else	if (reg_no == 12) {
						// LNA gain, MSbs
						a = sreg12;
					} else	if (reg_no == 13) {
						// RF gain, LSB
						a = RF_gain & 0x00FF;
						sreg14 = (RF_gain >> 8) & 0x00F;
					} else	if (reg_no == 14) {
						// RF gain, MSbs
						a = sreg14;
					} else	if (reg_no == 15) {
						// IF gain, LSB
						a = IF_gain & 0x00FF;
						sreg16 = (IF_gain >> 8) & 0x00F;
					} else	if (reg_no == 16) {
						// IF gain, MSbs
						a = sreg16;
					}
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999RGV%02X\r\n", a);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "GMS", 3)) {
					// new.
					// GMS command - Get multiple status registers (faster than asking for individual ones)
					// includes two parameters: number of registers to read (3 decimal digits) and start register (also 3 decimal digits)

					// number of registers to read
					c = rxLine.buf[7] - '0';
					c = c * 10 + rxLine.buf[8] - '0';
					c = c * 10 + rxLine.buf[9] - '0';
					nreg_no = c;
					// valid range is 0 - 128

					// starting at register number x
					c = rxLine.buf[10] - '0';
					c = c * 10 + rxLine.buf[11] - '0';
					c = c * 10 + rxLine.buf[12] - '0';
					start_reg_no = c;
					// valid range is 0 - 255

		 			// check length
		 			if((nreg_no > 128) || (start_reg_no > 255)){
		 				// return message would be too long or out of range
		 				snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999NAK\r\n");
		 			} else {
		 				snprintf((char *)txLine.buf, sizeof(txLine.buf),"@999MRV");	// Multiple Register Values
			 			for(i = 0; i< nreg_no; i++){
			 				reg_no = start_reg_no + i;
							a = 0;
							if (reg_no == 0) {
								// power good signals part 1
								c = power_check();
								sreg1 = (c >> 8) & 0x0FF;	// save for 2nd part query
								a = c & 0x0FF;
							} else	if (reg_no == 1) {
								// power good signals part 2
								a = sreg1;
							} else	if (reg_no == 2) {
								// RF synthesizer lock status
								if (synth_check_lock(1))
									a |= 0x01; // SREG2 bit 0
								if (synth_check_lock(2))
									a |= 0x02; // SREG2 bit 1

							} else	if (reg_no == 3) {
								// RSSI (RF_POWER_DET1), LSB
								a = aux_adc[0] & 0x00FF;
								sreg4 = (aux_adc[0] >> 8) & 0x00F;
							} else	if (reg_no == 4) {
								// RSSI (RF_POWER_DET1), MSbs
								a = sreg4;

							} else	if (reg_no == 5) {
								// RF power detect from the RF mixer (RF_POWER_DET2), LSB
								// save MSbs for later
								a = aux_adc[1] & 0x00FF;
								sreg6 = (aux_adc[1] >> 8) & 0x00F;
							} else	if (reg_no == 6) {
								// RF power detect from the RF mixer (RF_POWER_DET2), MSbs
								a = sreg6;

							} else	if (reg_no == 7) {
								// IF1 power detect (IF1_POWER_DET), LSB
								a = aux_adc[2] & 0x00FF;
								sreg8 = (aux_adc[2] >> 8) & 0x00F;
							} else	if (reg_no == 8) {
								// IF1 power detect (IF1_POWER_DET), MSbs
								a = sreg8;

							} else	if (reg_no == 9) {
								// Board temperature sensor, LSB
								a = aux_adc[4] & 0x00FF;
								sreg10 = (aux_adc[4] >> 8) & 0x00F;
							} else	if (reg_no == 10) {
								// Board temperature sensor, MSbs
								a = sreg10;

							} else	if (reg_no == 11) {
								// LNA gain (LSB)
								a = LNA_gain & 0x00FF;
								sreg12 = (LNA_gain >> 8) & 0x00F;
							} else	if (reg_no == 12) {
								// LNA gain, MSbs
								a = sreg12;
							} else	if (reg_no == 13) {
								// RF gain, LSB
								a = RF_gain & 0x00FF;
								sreg14 = (RF_gain >> 8) & 0x00F;
							} else	if (reg_no == 14) {
								// RF gain, MSbs
								a = sreg14;
							} else	if (reg_no == 15) {
								// IF gain, LSB
								a = IF_gain & 0x00FF;
								sreg16 = (IF_gain >> 8) & 0x00F;
							} else	if (reg_no == 16) {
								// IF gain, MSbs
								a = sreg16;
							}
	              			snprintf((char *)&txLine.buf[7 + 2*i],3,"%02X",a);
			 			}
              			snprintf((char *)&txLine.buf[7 + 2*i],3,"\r\n");
		 			}
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if ((!memcmp((const char *)&rxLine.buf[4], "GMR", 3)) || (!memcmp(
						(const char *)&rxLine.buf[4], "GMT", 3))) {
					// new. Get multiple registers (faster than asking for individual ones)
					// GMR command - get multiple registers (read from flash memory)
					// GMT command - get multiple registers temporary (read from volatile memory)

					// number of registers to read
					c = rxLine.buf[7] - '0';
					c = c * 10 + rxLine.buf[8] - '0';
					c = c * 10 + rxLine.buf[9] - '0';
					nreg_no = c;
					// valid range is 0 - 128

					// starting at register number x
					c = rxLine.buf[10] - '0';
					c = c * 10 + rxLine.buf[11] - '0';
					c = c * 10 + rxLine.buf[12] - '0';
					start_reg_no = c;
					// valid range is 0 - 255

		 			// check length
		 			if((nreg_no > 128) || (start_reg_no > 255)){
		 				// return message would be too long or out of range
		 				snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999NAK\r\n");
		 			} else {
		 				snprintf((char *)txLine.buf, sizeof(txLine.buf),"@999MRV");	// Multiple Register Values
			 			for(i = 0; i< nreg_no; i++){
							if (rxLine.buf[6] == 'R') {
								// read registers from flash memory
		              			snprintf((char *)&txLine.buf[7 + 2*i],3,"%02X",configSw.reg[start_reg_no+i]);
							} else {
								// read registers from volatile memory
		              			snprintf((char *)&txLine.buf[7 + 2*i],3,"%02X",regv[start_reg_no+i]);
							}
			 			}
              			snprintf((char *)&txLine.buf[7 + 2*i],3,"\r\n");
		 			}
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if ((!memcmp((const char *)&rxLine.buf[4], "GRG", 3)) || (!memcmp(
						(const char *)&rxLine.buf[4], "GRT", 3))) {
					// GRG command - get register (read from flash memory)
					// GRGT command - get register temporary (read from volatile memory)

					// check whether 1,2 or 3 digits
					c = rxLine.buf[7] - '0';
					if ((rxLine.buf[8] != '\r') && (rxLine.buf[8] != '\n')) {
						// 2 or 3 digit register number
						c = c * 10 + rxLine.buf[8] - '0';
					}
					if ((rxLine.buf[9] != '\r') && (rxLine.buf[9] != '\n')) {
						// 3 digit register number
						c = c * 10 + rxLine.buf[9] - '0';
					}
					reg_no = c;

					if ((reg_no < sizeof(configSw.reg)) && (rxLine.buf[6]
							== 'G'))
						a = configSw.reg[reg_no]; // GRG, read non-volatile register
					else if ((reg_no < sizeof(regv)) && (rxLine.buf[6] == 'T')) // make sure we do not write past the buffer end
						a = regv[reg_no]; // GRT, read volatile register
					else
						// trying to read non-volatile registers past index NREG_V
						a = 0;

					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999RGV%02X\r\n", a);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

//				} else if (!memcmp((const char *)&rxLine.buf[4], "EFS", 3)) {
//					// EFS command - Erase File Sector
//					// first check block number is valid
//					// flash sectors 17 through 29 (32KB*13 = 416KB) are reserved for one FPGA configuration
//					// Important: use the compress feature in Xilinx ISE to reduce file size.
//					if ((sect_no < 17) || (sect_no > 29)) {
//						retval = -1; // NAK
//					} else {
//						// make sure the block does not relate to a protected configuration
//						c = block_to_config(sect_no);
//						read_sw_config(c, &cs); // read the software configuration of interest from flash
//						if ((c == 0) && (PROTECTION == 1)) {
//							retval = -1; // boot configuration is protected
//						} else if (cs.protection) {
//							retval = -1; // configuration is protected
//						} else {
//							// IAP Erase Sector(s) command
//							retval = internalFlash_EraseSector(sect_no);
//						}
//					}
//					if (retval) {
//						// error -> NAK
//						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999NAK\r\n");
//					} else {
//						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999ACK\r\n");
//					}
//					txLine.wptr = strlen((const char *)txLine.buf); // set message length
//					sendLine(commandPortNo);
//
//				} else if (!memcmp((const char *)&rxLine.buf[4], "SCA", 3)) {
//					// SCA - Set Configuration start Address command
//					// read the selected configuration number for the specified start address
//					// decode selected configuration number
//					c = s78_decode - 1;
//					// make sure configuration pointer is within allowed boundaries
//					if (c < NSWCONFIGS) {
//						// decode 32-bit hex string, then save
//						l = 0;
//						l |= hex2bin((char *)&rxLine.buf[9]) & 0x000000FF;
//						l <<= 8;
//						l |= hex2bin((char *)&rxLine.buf[11]) & 0x000000FF;
//						l <<= 8;
//						l |= hex2bin((char *)&rxLine.buf[13]) & 0x000000FF;
//						l <<= 8;
//						l |= hex2bin((char *)&rxLine.buf[15]) & 0x000000FF;
//						if (c == active_sw_config) {
//							configSw.fpga_start_address = l;
//							// save the active configuration
//							configSw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
//							timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
//							save_configuration(c, &configSw); // save the hardware + active software configurations
//						} else {
//							// non-active configuration. read from flash
//							read_sw_config(c, &cs); // read the software configuration of interest from flash
//							cs.fpga_start_address = l;
//							// save this non-active configuration
//							timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
//							// Note: saving takes 120ms!! during which incoming messages are not processed
//							save_configuration(c, &cs); // save the hardware + active software + this non-active configurations
//						}
//					}
//
//				} else if (!memcmp((const char *)&rxLine.buf[4], "GCA", 3)) {
//					// GCA - Get Configuration start Address command
//					// read the selected configuration number for the specified start address
//					// decode selected configuration number
//					c = s78_decode - 1;
//					// make sure configuration pointer is within allowed boundaries
//					if (c < NSWCONFIGS) {
//						if (c == active_sw_config) {
//							// active configuration
//							l = configSw.fpga_start_address;
//						} else {
//							// non-active configuration. read from flash
//							read_sw_config(c, &cs); // read the software configuration of interest from flash
//							l = cs.fpga_start_address;
//						}
//						// Virtual location (not accounting bad-block redirection to another physical address)
//						if (l == 0xFFFFFFFF) {
//							// undefined
//							snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999CAS????????\r\n");
//						} else {
//							snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999CAS%08X\r\n", l);
//						}
//						txLine.wptr = strlen((const char *)txLine.buf); // set message length
//						sendLine(commandPortNo);
//					}
//
//				} else if (!memcmp((const char *)&rxLine.buf[4], "WFS", 3)) {
//					// WFS - Write File Sector command
//					// first check block number is valid
//					// flash sectors 17 through 29 (32KB*13 = 416KB) are reserved for one FPGA configuration
//					// Important: use the compress feature in Xilinx ISE to reduce file size.
//					if ((sect_no < 17) || (sect_no > 29)) {
//						retval = -1; // NAK
//					} else {
//						// make sure the block does not relate to a protected configuration
//						c = block_to_config(sect_no);
//						read_sw_config(c, &cs); // read the software configuration of interest from flash
//						if ((c == 0) && (PROTECTION == 1)) {
//							retval = -1; // boot configuration is protected
//						} else if (cs.protection) {
//							retval = -1; // configuration is protected
//						} else {
//							// disassemble the packet sequence number and data field
//							packet_no = 0;
//							for (i = 0; i < 4; i++)
//								packet_no = packet_no * 10 + rxLine.buf[i + 10]
//										- '0';
//							// convert the data field from hex ASCII to binary. Up to 128 bytes in the data field.
//							l = (rxLine.wptr - 16) / 2;
//							j = 0;
//							for (i = 14; i < rxLine.wptr - 2; i += 2) {
//								rxLine.buf[j++] = hex2bin((char *)&rxLine.buf[i]); // save binary in place
//							}
//
//							if (packet_no == 1) {
//								// first packet within a block. Reset the byte count.
//								rx_byte_count = 0;
//							}
//
//							if (packet_no == 9999) {
//								// last packet. Two cases: did we last end at a 256 byte boundary?
//								if ((rx_byte_count & 0x000000FF) == 0) {
//									// last write ended at a 256-byte boundary. No need for previous data in buf1 cache.
//									// fill empty space with FFs
//									memset((char *)&rxLine.buf[l], 0xFF, 256 - l);
//								} else {
//									// last write did not end at a 256-byte boundary.
//									// Insert previous data in buf1 cache.
//									memcpy((char *)&rxLine.buf[128], (const char *)&rxLine.buf[0],
//											128); // make room for 128 bytes
//									memcpy((char *)&rxLine.buf[0], (const char *)&buf1, 128); // insert previous 128 bytes
//									// fill in buffer with FFs if needed
//									if (l < 128) {
//										memset((char *)&rxLine.buf[128 + l], 0xFF, 128
//												- l);
//									}
//								}
//								// write 256 Bytes (min number) to internal flash
//								retval = internalFlash_write(sect_no,
//										(rx_byte_count & 0xFFFFFF00),
//										(char *)rxLine.buf, 256);
//								rx_byte_count += l;
//							} else if (((packet_no & 1) != 0) && (l == 128)) {
//								// odd packet number (not last one). Can't write to internal flash until we gather 256 Bytes
//								// Store until next WFS command
//								memcpy(&buf1, (const char *)rxLine.buf, l);
//								rx_byte_count += l;
//								retval = 0; // OK
//							} else if ((packet_no & 1) == 0) {
//								// even packet number
//								memcpy((char *)&rxLine.buf[128], (const char *)&rxLine.buf[0], 128); // make room for 128 bytes
//								memcpy((char *)&rxLine.buf[0], (const char *)&buf1, 128); // insert previous 128 bytes
//								// write size is limited to 256, 512, ... 4KB. No other values permitted.
//								// fill in buffer with FFs if needed
//								if (l < 128) {
//									memset((char *)&rxLine.buf[128 + l], 0xFF, 128 - l);
//								}
//								// write 256 Bytes (min number) to internal flash
//								retval = internalFlash_write(sect_no,
//										(rx_byte_count & 0xFFFFFF00),
//										(char *)rxLine.buf, 256);
//								rx_byte_count += l;
//							}
//						}
//					}
//
//					if (retval) {
//						// error -> NAK
//						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999NAK\r\n");
//					} else {
//						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999ACK%06d\r\n", rx_byte_count);
//					}
//					txLine.wptr = strlen((const char *)txLine.buf); // set message length
//					sendLine(commandPortNo);
//
//				} else if (!memcmp((const char *)&rxLine.buf[4], "GCS", 3)) {
//					// GCS - Get CheckSum command
//					retval = internalFlash_GCS(sect_no, (unsigned long *) (&c));
//
//					// first check block number is valid
//					if (retval) {
//						// block number out-of-range error -> NAK
//						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999NAK\r\n");
//					} else {
//						// return the 16-bit checksum
//						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999CKS%04X\r\n", c);
//					}
//					txLine.wptr = strlen((const char *)txLine.buf); // set message length
//					sendLine(commandPortNo);
//
//				} else if (!memcmp((const char *)&rxLine.buf[4], "PCF", 3)) {
//					// PCF - Protect Configuration command
//					// decode selected configuration number
//					c = s78_decode - 1;
//					// make sure configuration pointer is within allowed boundaries
//					if (c < NSWCONFIGS) {
//						if (c == active_sw_config) {
//							if (rxLine.buf[9] == 'U') {
//								// Unprotect configuration
//								configSw.protection = 0;
//							} else if (rxLine.buf[9] == 'P') {
//								// Protect configuration
//								configSw.protection = 1;
//							}
//							// save the active configuration
//							configSw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
//							timer[1] = TIMER1SLOW;
//							// re-arm timer[1]: wait before saving the comblock configuration
//							//into flash (to consolidate multiple successive changes into one sector programming operation)
//						} else {
//							// non-active configuration. read from flash
//							read_sw_config(c, &cs); // read the software configuration of interest from flash
//							if (rxLine.buf[9] == 'U') {
//								// Unprotect configuration
//								cs.protection = 0;
//							} else if (rxLine.buf[9] == 'P') {
//								// Protect configuration
//								cs.protection = 1;
//							}
//							// save this non-active configuration
//							timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
//							// Note: saving takes 120ms!! during which incoming messages are not processed
//							save_configuration(c, &cs); // save the hardware + active software + this non-active configurations
//						}
//					}
//
//				} else if (!memcmp((const char *)&rxLine.buf[4], "TST", 3)) {
//					// test test test
//					// read data at sector 17
//					c = 0x00018000 + 128 * s789_decode;
//					memcpy(&buf1[0], (void *) (c), sizeof(buf1));
//					for (i = 0; i < sizeof(buf1); i++) {
//						snprintf((char *)&txLine.buf[2 * i], 3, "%02X", buf1[i]);
//					}
//					snprintf((char *)&txLine.buf[2 * i], 3, "\r\n");
//					txLine.wptr = strlen((const char *)txLine.buf); // set message length
//					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "RST", 3)) {
					// RST command - reset
					goto rst;
					/*
					 // reset micro
					 ResetISR();
					 */

				}

				// end of message processing section
				rxLine.rptr = rxLine.wptr; // release rx message


			}
		}
	}
	// end of main loop ------------------------------------------------------------------------


	return (0);
}

/*** EOF ***/
